#!/usr/bin/perl

use v5.10;
use strict;
use warnings;
binmode STDOUT, ":utf8";

use XML::Parser;
use HTML::Entities;
use Getopt::Std;
use File::Basename;
#use Data::Dumper;

my $version = "1.0";

my %mcfiles = ();

# -d: debug
my %opts;
my $opt_letters = "dv";

my $prog = basename($0);

getopts($opt_letters, \%opts);

die "Usage: $prog [-$opt_letters]  mplfile1 mplfile2 [...]\nStopped"
    unless @ARGV >= 2;

sub debug;
sub print_mpl;

for (@ARGV) {
    my $file = $_;

    my $parser = new XML::Parser(ErrorContext => 2);

    $parser->setHandlers(
	Char =>		\&char_handler,
	Start =>	\&start_handler,
	End =>		\&end_handler,
	Final =>	\&final_handler,
	Default =>	\&default_handler
    );

    $parser->parsefile($file);
    $parser = undef;

}
print_mpl '<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>';
print_mpl "<MPL Version=\"2.0\" Title=\"Library Merge @ARGV\">";
print_mcfiles();
print_mpl '</MPL>';

# end

# handlers below

my @paths;		# xml paths
my %item;
my $current_fieldname;

sub start_handler {
    my ($p, $el) = @_;

    push @paths, my $path = join('::', $p->context, $el);

    sub get_values {
	my $attr_re = shift;
	my $ret = '';

	while (@_) {
	    my ($a, $v) = (shift, shift);
	    $ret = ($ret ? join($1 ? '=' : ' ', $ret, $v) : $v)  if $v and $a =~ /$attr_re/;
	}
	return $ret;
    }

    debug "PATH: ", $path;
    if (defined $p->current_element) {

	if ($path eq 'MPL::Item::Field') {
	    my $val = get_values('^(?:Name)$', @_);
	    debug "\tField Name: ", $val;
	    $current_fieldname = $val;
	}

	elsif ($path eq 'MPL::Item') {
	    %item = ();
	}
    }
}

sub char_handler {
    my ($p, $data) = @_;

    my $path = $paths[-1];

    if ($current_fieldname) {
	# the expat parser returns entities as single characters
	if ($data eq '&' or $data eq '<' or $data eq '>') { $data = encode_entities($data); }
	else					      { $data = decode_entities($data); }

	if ($data eq "\n") {
	    $item{$current_fieldname} .= "\r\n";
	}
	else{
	    $item{$current_fieldname} .= $data;
	}
    }
}

sub end_handler {
    my ($p, $el) = @_;

    sub max ($$) { $_[$_[0] < $_[1]] }

    my $path = pop @paths;

    debug "End: ", $path;
    if ($path eq 'MPL::Item::Field') {
	debug "\tField Value: ", $item{$current_fieldname};
	$current_fieldname = undef;
    }
    elsif ($path eq 'MPL::Item') {

	my $f = $item{'Filename'};
	if (!exists $mcfiles{$f}) {
	    debug 'NEW: adding all fields: ', $f;
	    map { $mcfiles{$f}{$_} = $item{$_} } keys %item;
	}
	else {
	    while (my ($k, $v) = each %item) {
		if ($k eq 'Number Plays' or $k eq 'Skip Count') {
		    $mcfiles{$f}{$k} += $v;
		    debug "Updating $f: $k = $mcfiles{$f}{$k}";
		}
		elsif ($k eq 'Last Played' or
		       $k eq 'Last Skipped' or
		       $k eq 'Date Created' or
		       $k eq 'Date Modified' or
		       $k eq 'Date Imported') {
		    $mcfiles{$f}{$k} = max($mcfiles{$f}{$k}, $v);
		    debug "Updating $f: $k = $mcfiles{$f}{$k}";
		}
		else {
		    if (!exists $mcfiles{$f}{$k} or $item{'Date Modified'} > $mcfiles{$f}{'Date Modified'}) {
			debug "Updating: $f: $k = $v";
			$mcfiles{$f}{$k} = $v;
		    }
		}
	    }
	}
	%item = ();
    }
}

sub final_handler {
    #print Dumper(\%mcfiles);
}

sub default_handler {
}

sub print_mpl {
    local $\ = "\r\n";

    print @_;

}

sub print_mcfiles {
    local $\ = "\r\n";

    for my $file (keys %mcfiles) {
	print '<Item>';
	for my $key (keys $mcfiles{$file}) {
	    print '<Field Name="', $key, '">', $mcfiles{$file}{$key}, '</Field>';
	}
	print '</Item>';
    }
}

sub debug {
    return unless $opts{'d'};

    say((split(/::/, (caller(1))[3] || 'main'))[-1], ': ', @_);
}
